/*
 * lm2latex - src/l2l.c
 * Copyright (C) 2010 Sebastian Harl <sh@tokkee.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * A Linux-Magazin markup to LaTeX converter -- core library.
 */

#include "l2l.h"
#include "l2l_scanner.h"
#include "l2l_features.h"

#include <assert.h>

#include <stdio.h>

/*
 * private helper functions
 */

static void
write_tex_preamble(FILE *out)
{
	assert(out);

	fprintf(out,
"\\documentclass[11pt,a4paper,twoside,twocolumn]{article}\n"
"\n"
"\\usepackage{ifpdf}\n"
"\n"
"\\usepackage{a4}\n"
"\\usepackage{fancyhdr}\n"
"\\ifpdf\n"
"\\usepackage[pdftex]{graphicx}\n"
"\\else\n"
"\\usepackage[dvips]{graphicx}\n"
"\\fi\n"
"\\usepackage[svgnames]{xcolor}\n"
"\n"
"\\usepackage{listings}\n"
"\\lstset{breaklines=true,breakatwhitespace=true}\n"
"\\lstset{prebreak=\\mbox{$\\hookleftarrow$}}\n"
"\\lstset{basicstyle=\\footnotesize\\ttfamily}\n"
"\n"
"\\usepackage[ngerman]{babel}\n"
"\\usepackage[latin1]{inputenc}\n"
"\\usepackage[T1]{fontenc}\n"
"\\usepackage{lmodern}\n"
"\n"
"\\ifpdf\n"
"\\usepackage[a4paper,margin=15mm,top=25mm,bottom=20mm,pdftex]{geometry}\n"
"\\else\n"
"\\usepackage[a4paper,margin=15mm,top=25mm,bottom=20mm,dvips]{geometry}\n"
"\\fi\n"
"\n"
"\\usepackage{caption}\n"
"\\usepackage{float}\n"
"\\usepackage{url}\n"
"\n"
"\\floatstyle{ruled}\n"
"\\newfloat{l2lbox}{tbp}{lof}\n"
"\n"
"\\setlength{\\parskip}{6pt}\n"
"\n"
"\\renewcommand{\\captionfont}{\\footnotesize}\n"
"\\setcounter{secnumdepth}{-1}\n"
"\n"
"\\renewcommand{\\headrulewidth}{0.4pt}\n"
"\\renewcommand{\\footrulewidth}{0pt}\n"
"\n"
"\\fancypagestyle{headings}\n"
"\\fancyhead{}\n"
"\\fancyhead[LO,RE]{}\n"
"\\fancyhead[LE,RO]{\\thepage}\n"
"\\fancyfoot{}\n"
"\n"
"\\pagestyle{fancy}\n"
"\n");
} /* write_tex_preamble */

/*
 * public API
 */

int
l2l_convert_file_fh(FILE *in, FILE *out)
{
	yyscan_t scanner;

	int status;

	write_tex_preamble(out);

	l2l_yylex_init(&scanner);
	l2l_yyset_in(in, scanner);
	l2l_yyset_out(out, scanner);

	status = l2l_yylex(scanner);

	l2l_yylex_destroy(scanner);
	return status;
} /* l2l_convert_file_fh */

int
l2l_convert_file(const char *infile, const char *outfile)
{
	FILE *in, *out;

	int status;

	in = fopen(infile, "r");
	if (! in)
		return -1;

	out = fopen(outfile, "w");
	if (! out) {
		fclose(in);
		return -1;
	}

	status = l2l_convert_file_fh(in, out);
	fclose(in);
	fclose(out);
	return status;
} /* l2l_convert_file */

unsigned int
l2l_version(void)
{
	return L2L_VERSION;
} /* l2l_version */

const char *
l2l_version_string(void)
{
	return L2L_VERSION_STRING;
} /* l2l_version_string */

const char *
l2l_version_extra(void)
{
	return L2L_VERSION_EXTRA;
} /* l2l_version_extra */

/* vim: set tw=78 sw=4 ts=4 noexpandtab : */

